## plot a numeric vector on the map.  Polygon names are taken from names(x).
## Note: use the correct encoding (e.g. latin-1) if the names include 8-bit
## symbols. 
mapPlot <- function(x, database, main=NULL,
                    col=topo.colors(100), ...) {
### x - numeric vector, names correspond to the map regions
### database - map database
### main - title of the map
    ## plot all polygons which match name with color col
    mkPlot <- function(name, col) {
        map(database, name, col=col, fill=T, add=T)
    }
    ## calculate color using the value for region
    mkCol <- function(val) {
        i <- as.integer((Ncol - 1)*(val - x0)/(x1 - x0))
        i <- ifelse(i < 1, 1, i)
        i <- ifelse(i > Ncol, Ncol, i)
        col[i]
    }
    opar <- par(xpd=TRUE)
    on.exit(par(opar), add=TRUE)
    Ncol <- length(col)
    x0 <- range(x)[1]
    x1 <- range(x)[2]
    r <- map(database, ...)$range
    if(!is.null(main))
        mtext(main, line=1, cex=1.5)
    sapply(seq(along=x), FUN=function(i) mkPlot(names(x[i]), mkCol(x[i])))
    lev <- pretty(x, n=10)
    legx <- r[2]
    legy <- r[4]
    legend(legx, legy, legend=as.character(lev), fill=mkCol(lev),
           bty="n")
    invisible(NULL)
}
